/**************************************************
* tcp_func.h  V2.1 - Header fuer Windows (VS) UND Linux
* (C) JoEmbedded 08.08.2022
***************************************************/

#define _WINSOCK_DEPRECATED_NO_WARNINGS
#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <assert.h>

#include "tcp_func.h"    // Jo's W32/Linux-Wrapper. Immer als ERSTER eigener Header

#define emb_log_printf printf

#ifdef _MSC_VER
// ----------- Sockets -----------------------
// WIN32 Init Windows Spezial Winsock, 0: OK, sonst Error
void tf_init_tcp(void){
    WSADATA wsaData;
    int iResult;
    iResult = WSAStartup(MAKEWORD(2,2), &wsaData);      // wsaData wird Versionsnummer
    if (iResult != 0) {
        emb_log_printf("WSAStartup() failed with error: %d\n", iResult);
        exit(-1);
    }
}

// WIN32 Gegenstueck zu Init
void tf_exit_tcp(void){
    WSACleanup();
}


// Fehlercode holen
int tf_lasterror_no(void){
        return WSAGetLastError();
}

#else

// ----------- Sockets -----------------------
// WIN32 Init Windows Spezial Winsock, 0: OK, sonst Error
void tf_init_tcp(void){
	;
}

// WIN32 Gegenstueck zu Init
void tf_exit_tcp(void){
	;
}


// Fehlercode holen
int tf_lasterror_no(void){
	return errno;
}
#endif




// hostname, IP oder NULL, port 0-65535 oder service, Fehler: 0
struct addrinfo* tf_init_addrinfo(char *hostname,char *service){
    int res;
    struct addrinfo *result = NULL;         // Zeigt auf was wir bekommen
    struct addrinfo hints;                  // Hints geben an, wa ir gerne haetten

    ZeroMemory(&hints, sizeof(hints));      // Systemfkt. Windows
    hints.ai_family = AF_INET;              // Hints mit Standardwerten belegen
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_flags = AI_PASSIVE;

    // Resolve the server address and port
    res = getaddrinfo(/* Ziel-IP/Domaene*/ hostname, /* Ziel-Port*/ service, &hints, &result); // mappt hints->result...
    if ( res != 0 ) {
        emb_log_printf("getaddrinfo() failed with error: %d\n", res);
    }
    return result;
}

// Socket generieren - Wir gehen davon aus, dass das immer funktioniert, also kein Error
SOCKET tf_socket(struct addrinfo *paddrinfo){
    SOCKET lsock;

    assert(paddrinfo);

    lsock = socket(paddrinfo->ai_family, paddrinfo->ai_socktype, paddrinfo->ai_protocol);
#ifdef _MSC_VER
    if (lsock == INVALID_SOCKET) {
#else
    if (lsock < 0) {
#endif

        emb_log_printf("socket() failed with error: %ld\n", tf_lasterror_no());
        freeaddrinfo(paddrinfo);
        tf_exit_tcp();
        exit(-1);
    }
    // Momentan noch Abbruch bei Fehler, spaerer evtl. NULL zurueckgeben
    return lsock;
}


// bind
void tf_bind(SOCKET lsock, struct addrinfo *paddrinfo){
    int res;

    res = bind( lsock, paddrinfo->ai_addr, (int)paddrinfo->ai_addrlen);
#ifdef _MSC_VER
    if (res == SOCKET_ERROR) {
#else
    if (res < 0) {
#endif
        emb_log_printf("bind() failed with error: %d (Port not available?)\n", tf_lasterror_no());
        freeaddrinfo(paddrinfo);
        closesocket(lsock);
        tf_exit_tcp();
        exit(-1);
    }
    freeaddrinfo(paddrinfo);
}

/**** Wichtige Routine, bei der schiefgehen mit 'schiefgehen' gerechnet werden muss ****/
int tf_connect(SOCKET lsock,  struct addrinfo *paddrinfo){
    int res;

    res = connect(lsock, paddrinfo->ai_addr, paddrinfo->ai_addrlen);

    /****** Wird uebergeordnet abgefangen
    if (res == SOCKET_ERROR oder <0 ) {
        emb_log_printf("bind() failed with error: %d\n", tf_lasterror_no());
        freeaddrinfo(paddrinfo);
        closesocket(lsock);
        tf_exit_tcp();
        exit(-1);
    }
    ******/

    freeaddrinfo(paddrinfo);
    return res;
}


void tf_listen(SOCKET lsock, int maxconn){ // maxconn (Tiefe der Anfrageque) oder 0 fuer 'vernuenftigen Wert'
    int res;
    res = listen(lsock, maxconn?maxconn:SOMAXCONN);  // Angabe Maximal Nummer oder Systemkonstante
#ifdef _MSC_VER
    if (res == SOCKET_ERROR) {
#else
    if (res < 0) {
#endif

        emb_log_printf("listen() failed with error: %d\n", tf_lasterror_no());
        closesocket(lsock);
        tf_exit_tcp();
        exit(-1);
    }
}

SOCKET tf_accept(SOCKET lsock, struct sockaddr_in *premoaddr){
   SOCKET csock;
   int lerr;
   int plen=sizeof(struct sockaddr_in);

   // premoaddr 0 oder gueltiger Zeiger
   csock = accept(lsock, (struct sockaddr*)premoaddr,premoaddr==NULL?0:&plen );
#ifdef _MSC_VER
   if (csock == INVALID_SOCKET) {
#else
   if (csock < 0) {
#endif
        lerr=tf_lasterror_no();
        if(lerr!=WSAEINTR /* 10004 */ && lerr!= WSAEINVAL ){
                emb_log_printf("accept() failed with error: %d\n", lerr);
        }
        closesocket(lsock);
        return 0;       // NIX
    }
    return csock;
}

// ---- res<=0: Close, >0: Anz, Bytes
/**** Wichtige Routine, bei der schiefgehen mit 'schiefgehen' gerechnet werden muss ****/
int tf_recv(SOCKET rsocket, char* read_buf, int rsize){
   int res;
   res = recv(rsocket, read_buf, rsize, 0); // 0: Flag: Peek, Waitall oder OutOfBand Data

   /***** Wird auch ausgeloest durch Medlungen wie z.B. Connection Reset by Peer (10054), was ja OK ist...
   if(res<0){
        emb_log_printf("recv failed with error: %d\n", tf_lasterror_no());
        closesocket(rsocket);
        tf_exit_tcp();
        exit(-1);
   }
   ******/
   return res;
}

/**** Wichtige Routine, bei der schiefgehen mit 'schiefgehen' gerechnet werden muss ****/
int tf_send(SOCKET ssocket, char* write_buf, int wsize){
   int res;
#ifdef _MSC_VER
   // Unter Windows gibts MSG_NOSIGNAL nicht
   res = send(ssocket, write_buf, wsize, 0);
#else
   //  MSG_NOSIGNAL ist wichtig unter LINUX, sonst terminiert sich das Prg!
   res = send(ssocket, write_buf, wsize, MSG_NOSIGNAL);
#endif
   if(res!=wsize) {
        /*
                printf("\n\n*** ACHTUNG: SOll/Ist: %d/%d *****\n\n",wsize,res);
                MessageBeep(-1);
        */
        res=-1;       // Es wurd nicht alles verschickt, dann Res def. <0
   }
   return res;
}


int tf_closesocket(SOCKET lsock){
#ifdef _MSC_VER
        shutdown(lsock,SD_BOTH);
#else
        shutdown(lsock,2);
#endif
        return closesocket(lsock);
}

// Warte auf bereiten Socket Daten oder Timeout von EINEM Socket
int tf_rdselect1(SOCKET lsock,  struct timeval* timeout){
#ifdef _MSC_VER
        int res;
        fd_set fdSetRead;
        FD_ZERO(&fdSetRead);    // select-Liste jedesmal neu erzeugen
        FD_SET(lsock,&fdSetRead);
        res= select(0, &fdSetRead, NULL, NULL, timeout); // W32 Par 0 obsolete
#else
        int res;
		TIMEVAL ltimeout;		// Linux: Manche OS modifizieren das Timeout (z.B. Raspi). Unter Windows; Timeout nicht modifiziert
        fd_set fdSetRead;
        FD_ZERO(&fdSetRead);    // select-Liste jedesmal neu erzeugen
        FD_SET(lsock,&fdSetRead);
		ltimeout=*timeout;		// Lokale Timeout Kopie herstellen
        res= select(lsock+1, &fdSetRead, NULL, NULL, &ltimeout);   // Linux benoetigt Par. 0
#endif
        return res;
}

// -------- Komplexe Funktion um den Serverpot zu oeffnen (als Port sind auch Namen erlaubt, wie http...) --------
SOCKET tf_create_server_socket(char *server_port, int maxclients){
    SOCKET lsock;
    struct addrinfo *paddrinfo;

    // Serverport initialisieren/mappen/konvertieren
    //                         kein Host, aber Port gg.
    paddrinfo = tf_init_addrinfo(NULL,server_port);
    if(!paddrinfo) 
    {
	return (SOCKET)(0);
    }

    // socket() - create a SOCKET for connecting to server
    lsock = tf_socket(paddrinfo);

    // bind() - Socket belegen
    tf_bind(lsock, paddrinfo);

    // listen()
    tf_listen(lsock,maxclients);

    return lsock;
}


SOCKET tf_create_client_socket(char* psn, char* psp){
    SOCKET lsock;
    struct addrinfo *paddrinfo;
    int res;

    // Serverport initialisieren/mappen/konvertieren
    //                         kein Host, aber Port gg.
    paddrinfo = tf_init_addrinfo(psn,psp);
    if(!paddrinfo) 
    {
	return (SOCKET)(0);
    }

    // socket() - create a SOCKET for connecting to server
    lsock = tf_socket(paddrinfo);

    res=tf_connect(lsock, paddrinfo);
#ifdef _MSC_VER
    if(res == SOCKET_ERROR) {
#else
    if(res < 0) {
#endif
        tf_closesocket(lsock);
	return (SOCKET)(0);
    }

    return lsock;
}

// END

